/*-
 * Copyright (c) 2011 - 2014 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Rozhuk Ivan <rozhuk.im@gmail.com>
 *
 */


#include <sys/param.h>

#ifdef __linux__ /* Linux specific code. */
#define _GNU_SOURCE /* See feature_test_macros(7) */
#define __USE_GNU 1
#endif /* Linux specific code. */

#include <sys/types.h>
#include <net/if.h>

#include <errno.h>
#include <inttypes.h>
#include <stdio.h> /* snprintf, fprintf */
#include <string.h> /* bcopy, bzero, memcpy, memmove, memset, strerror... */
#include <stdlib.h> /* malloc, exit */
#include <unistd.h> /* close, write, sysconf */
#include <fcntl.h> /* open, fcntl */
#include <signal.h> /* SIGNAL constants. */

#include "mem_find.h"
#include "buf_case.h"
#include "StrToNum.h"
#include "xml.h"

#include "core_macro.h"
#include "core_helpers.h"
#include "core_net_helpers.h"
#include "core_log.h"
#include "core_upnp_ssdp.h"



#include <config.h>
#undef PACKAGE_NAME
#define PACKAGE_NAME	"SSDP announce daemon for UPnP 1.1"
#define CFG_FILE_MAX_SIZE (1024 * 1024)


struct prog_settings {
	thrp_p		thrp;
	upnp_ssdp_p	upnp_ssdp;	/* UPnP SSDP server. */
	int		log_fd;		// log file descriptor
	cmd_line_data_t	cmd_line_data;	/*  */
};
static struct prog_settings g_data;


static void	SigInstall(void);
static void	SigHandler(int iSigNum);

#define SSDPD_CFG_CALC_VAL_COUNT(args...)					\
	xml_calc_tag_count_args(cfg_file_buf, cfg_file_buf_size,		\
	    (const uint8_t*)"ssdpd", ##args)
#define SSDPD_CFG_GET_VAL_DATA(next_pos, data, data_size, args...)		\
	xml_get_val_args(cfg_file_buf, cfg_file_buf_size, next_pos,		\
	    NULL, NULL, data, data_size, (const uint8_t*)"ssdpd", ##args)
#define SSDPD_CFG_GET_VAL_INT(next_pos, val_ret, args...)			\
	xml_get_val_int_args(cfg_file_buf, cfg_file_buf_size, next_pos,		\
	    (int32_t*)val_ret, (const uint8_t*)"ssdpd", ##args)
#define SSDPD_CFG_GET_VAL_SSIZE(next_pos, val_ret, args...)			\
	xml_get_val_ssize_t_args(cfg_file_buf, cfg_file_buf_size, next_pos,	\
	    (ssize_t*)val_ret, (const uint8_t*)"ssdpd", ##args)


int
main(int argc, char *argv[]) {
	int error;

	memset(&g_data, 0, sizeof(g_data));
	g_data.log_fd = -1;
	if (0 != cmd_line_parse(argc, argv, &g_data.cmd_line_data)) {
		cmd_line_usage(PACKAGE, VERSION);
		return (0);
	}
	if (g_data.cmd_line_data.verbose) {
		/*g_data.log_fd = open("/root/trp/src/trp.log",
		    (O_WRONLY | O_APPEND | O_CREAT), 0777);//*/
		g_data.log_fd = open("/dev/stdout", (O_WRONLY | O_APPEND));
		core_log_fd = g_data.log_fd;
	}
	error = thrp_create(0, 0, 1, &g_data.thrp);
	if (0 != error) {
		LOG_ERR(error, "thrp_create()");
		return (error);
	}

	log_write("\n\n\n\n", 4);
	LOG_INFO(PACKAGE_NAME" "PACKAGE_VERSION": started");
	LOG_INFO_FMT("Build: "__DATE__" "__TIME__", "
#ifdef DEBUG
	    "DEBUG"
#else
	    "Release"
#endif
	);
	LOG_INFO_FMT("CPU count: %d", get_cpu_count());
	LOG_INFO_FMT("descriptor table size: %d (max files)", getdtablesize());

    { // process config file
	uint8_t *data, *ptm, *cur_pos, *cur_svc_pos, *cur_if_pos, *ann_data, *svc_data;
	uint8_t *cfg_file_buf, *cfg_dev_buf;
	size_t cfg_file_buf_size, cfg_dev_buf_size;
	size_t tm, data_size, ann_data_size, svc_data_size;
	uint8_t *uuid, *domain_name, *type;
	size_t domain_name_size, type_size;
	uint32_t ver, config_id, max_age, ann_interval;
	upnp_ssdp_dev_p dev;
	uint8_t *if_name, *url4, *url6;
	size_t if_name_size, url4_size, url6_size;
	upnp_ssdp_settings_t ssdp_st;
	uint8_t url4_buf[1024], url6_buf[1024];
	struct sockaddr_storage addr;


	error = read_file(g_data.cmd_line_data.cfg_file_name, 0, CFG_FILE_MAX_SIZE,
	    &cfg_file_buf, &cfg_file_buf_size);
	if (0 != error) {
		LOG_ERR(error, "config read_file()");
		return (error);
	}
	/* SSDP receiver. */
	if (0 == SSDPD_CFG_CALC_VAL_COUNT("announceList", "announce", NULL)) {
		LOG_INFO("no announce devices specified, nothink to do...");
		return (0);
	}
	/* Default values. */
	upnp_ssdp_def_settings(&ssdp_st);
	/* Read from config. */
	SSDPD_CFG_GET_VAL_INT(NULL, &ssdp_st.skt_rcv_buf, "skt", "rcvBuf", NULL);
	SSDPD_CFG_GET_VAL_INT(NULL, &ssdp_st.skt_snd_buf, "skt", "sndBuf", NULL);
	SSDPD_CFG_GET_VAL_INT(NULL, &ssdp_st.ttl, "skt", "ttl", NULL);
	SSDPD_CFG_GET_VAL_INT(NULL, &ssdp_st.hop_limit, "skt", "hopLimit", NULL);
	ssdp_st.flags = 0;
	if (0 == SSDPD_CFG_GET_VAL_DATA(NULL, NULL, NULL,
	    "flagEnableIPv4", NULL))
		ssdp_st.flags |= UPNP_SSDP_S_F_IPV4;
	if (0 == SSDPD_CFG_GET_VAL_DATA(NULL, NULL, NULL,
	    "flagEnableIPv6", NULL))
		ssdp_st.flags |= UPNP_SSDP_S_F_IPV6;
	if (0 == SSDPD_CFG_GET_VAL_DATA(NULL, &data, &data_size,
	    "httpServer", NULL) &&
	    0 != data_size &&
	    (sizeof(ssdp_st.http_server) - 1) > data_size) {
		ssdp_st.http_server_size = (uint32_t)data_size;
		memcpy(ssdp_st.http_server, data, data_size);
	}
	/* Create SSDP receiver. */
	error = upnp_ssdp_create(g_data.thrp, &ssdp_st, &g_data.upnp_ssdp);
	if (0 != error) {
		LOG_ERR(error, "upnp_ssdp_create()");
		return (error);
	}

	/* Add Devices to announce on selected ifaces. */
	cur_pos = NULL;
	while (0 == SSDPD_CFG_GET_VAL_DATA(&cur_pos, &ann_data, &ann_data_size,
	    "announceList", "announce", NULL)) {
		if (0 == xml_calc_tag_count_args(ann_data, ann_data_size,
		    (const uint8_t*)"ifList", "if", NULL))
			continue; /* No network interfaces specified. */
		if (0 != xml_get_val_args(ann_data, ann_data_size, NULL,
		    NULL, NULL, &data, &data_size, (const uint8_t*)"xmlDevDescr", NULL))
			continue; /* No xml file with UPnP dev description. */
		error = read_file((const char*)data, data_size, CFG_FILE_MAX_SIZE,
		    &cfg_dev_buf, &cfg_dev_buf_size);
		if (0 != error) {
			LOG_ERR(error, "xmlDevDescr read_file()");
			continue;
		}
		/* Load device options and add. */
		if (0 != xml_get_val_args(cfg_dev_buf, cfg_dev_buf_size, NULL,
		    NULL, NULL, (uint8_t**)&uuid, &tm,
		    (const uint8_t*)"root", "device", "UDN", NULL) ||
		    (5 + 36) != tm) { /* 5 = "uuid:" */
			LOG_ERR(EINVAL, "Invalid device UUID");
			free(cfg_dev_buf);
			continue;
		}
		uuid += 5; /* Skeep "uuid:". */
		if (0 != xml_get_val_args(cfg_dev_buf, cfg_dev_buf_size, NULL,
		    NULL, NULL, &data, &data_size,
		    (const uint8_t*)"root", "device", "deviceType", NULL)) {
no_dev_type:
			LOG_ERR(EINVAL, "No deviceType");
			free(cfg_dev_buf);
			continue;
		}
		/* Parce: "urn:schemas-upnp-org:device:MediaServer:3". */
		domain_name = (data + 4); /* Skeep "urn:". */
		ptm = mem_find_byte(5, data, data_size, ':');
		if (NULL == ptm)
			goto no_dev_type;
		domain_name_size = (ptm - domain_name);
		type = (ptm + 8); /* Skeep ":device:". */
		ptm = mem_find_byte((type - data), data, data_size, ':');
		if (NULL == ptm)
			goto no_dev_type;
		type_size = (ptm - type);
		ptm += 1; /* Skeep ":". */
		ver = UStr8ToUNum32(ptm, (data_size - (ptm - data)));
		
		config_id = 1; /* XXX not read. */
		max_age = 0;
		ann_interval = 0;
		xml_get_val_int_args(ann_data, ann_data_size, NULL,
		    (int32_t*)&max_age, (const uint8_t*)"maxAge", NULL);
		xml_get_val_int_args(ann_data, ann_data_size, NULL,
		    (int32_t*)&ann_interval, (const uint8_t*)"interval", NULL);
		/* Add device. */
		error = upnp_ssdp_dev_add(g_data.upnp_ssdp, (const char*)uuid,
		    (const char*)domain_name, domain_name_size, 
		    (const char*)type, type_size, ver,
		    time(NULL), config_id, max_age, ann_interval, &dev);
		if (0 != error) {
			LOG_ERR(error, "upnp_ssdp_dev_add()");
			free(cfg_dev_buf);
			continue;
		}
		/* Load services options and add. */
		cur_svc_pos = NULL;
		while (0 == xml_get_val_args(cfg_dev_buf, cfg_dev_buf_size, &cur_svc_pos, 
		    NULL, NULL, &svc_data, &svc_data_size,
		    (const uint8_t*)"root", "device", "serviceList", "service", NULL)) {
			if (0 != xml_get_val_args(svc_data, svc_data_size, NULL,
			    NULL, NULL, &data, &data_size,
			    (const uint8_t*)"serviceType", NULL)) {
no_svc_type:
				LOG_ERR(EINVAL, "No serviceType");
				continue;
			}
			/* Parce: "urn:schemas-upnp-org:service:ContentDirectory:3". */
			domain_name = (data + 4); /* Skeep "urn:". */
			ptm = mem_find_byte(5, data, data_size, ':');
			if (NULL == ptm)
				goto no_svc_type;
			domain_name_size = (ptm - domain_name);
			type = (ptm + 9); /* Skeep ":service:". */
			ptm = mem_find_byte((type - data), data, data_size, ':');
			if (NULL == ptm)
				goto no_svc_type;
			type_size = (ptm - type);
			ptm += 1; /* Skeep ":". */
			ver = UStr8ToUNum32(ptm, (data_size - (ptm - data)));
			/* Add service to device. */
			error = upnp_ssdp_svc_add(dev,
			    (const char*)domain_name, domain_name_size, 
			    (const char*)type, type_size, ver);
			if (0 != error) {
				LOG_ERR(error, "upnp_ssdp_svc_add()");
				continue;
			}
		}
		free(cfg_dev_buf);
		/* UPnP Device and services added, now add network interfaces. */
		cur_if_pos = NULL;
		while (0 == xml_get_val_args(ann_data, ann_data_size, &cur_if_pos, 
		    NULL, NULL, &data, &data_size,
		    (const uint8_t*)"ifList", "if", NULL)) {
			url4_size = 0;
			url6_size = 0;
			xml_get_val_args(data, data_size, NULL, NULL, NULL,
			    &url4, &url4_size, (const uint8_t*)"DevDescrURL4", NULL);
			xml_get_val_args(data, data_size, NULL, NULL, NULL,
			    &url6, &url6_size, (const uint8_t*)"DevDescrURL6", NULL);
			if (0 != xml_get_val_args(data, data_size, NULL, NULL, NULL,
			    &if_name, &if_name_size, (const uint8_t*)"ifName", NULL) ||
			    (0 == url4_size && 0 == url6_size)) {
				LOG_ERR(EINVAL, "Bad device network interface");
				continue;
			}
			/* Autoreplace NULL addr to if addr. */
			if (14 < url4_size && /* http://0.0.0.0... */
			    0 == buf_cmp((url4 + 7), 7, "0.0.0.0", 7) &&
			    (':' == url4[14] || '\\' == url4[14])) {
				LOG_EV("Autoreplace NULL IPv4 addr to if addr.");
				error = get_if_addr_by_name((const char*)if_name,
				    if_name_size, AF_INET, &addr);
				if (0 != error) {
					LOG_ERR(error, "get_if_addr_by_name()");
					continue;
				}
				error = ss_to_str_addr(&addr, (char*)(url4_buf + 7),
				    (sizeof(url4_buf) - 8), &tm);
				if (0 != error) {
					LOG_ERR(error, "ss_to_str_addr()");
					continue;
				}
				if (sizeof(url4_buf) > (url4_size + (tm - 7))) {
					memcpy(url4_buf, "http://", 7);
					memcpy((url4_buf + 7 + tm),
					    (url4 + 14),
					    (url4_size - 14));
					url4 = url4_buf;
					url4_size += (tm - 7);
					url4[url4_size] = 0;
					LOG_EV((char*)url4);
				} else {
					LOG_EV("URL to long, not enough buf space.");
					url4 = NULL;
					url4_size = 0;						
				}
			}
			if (13 < url6_size && /* http://[::]... */
			    0 == buf_cmp((url6 + 7), 4, "[::]", 4) &&
			    (':' == url6[11] || '\\' == url6[11])) {
				LOG_EV("Autoreplace NULL IPv6 addr to if addr.");
				error = get_if_addr_by_name((const char*)if_name,
				    if_name_size, AF_INET6, &addr);
				if (0 != error) {
					LOG_ERR(error, "get_if_addr_by_name()");
					continue;
				}
				error = ss_to_str_addr(&addr, (char*)(url6_buf + 7),
				    (sizeof(url6_buf) - 8), &tm);
				if (0 != error) {
					LOG_ERR(error, "ss_to_str_addr()");
					continue;
				}
				if (sizeof(url6_buf) > (url6_size + (tm - 7))) {
					memcpy(url6_buf, "http://", 7);
					memcpy((url6_buf + 7 + tm),
					    (url6 + 11),
					    (url6_size - 11));
					url6 = url6_buf;
					url6_size += (tm - 4);
					url6[url6_size] = 0;
					LOG_EV((char*)url6);
				} else {
					LOG_EV("URL to long, not enough buf space.");
					url6 = NULL;
					url6_size = 0;						
				}
			}
			/* Add service to device. */
			error = upnp_ssdp_dev_if_add(g_data.upnp_ssdp, dev,
			    (const char*)if_name, if_name_size,
			    (const char*)url4, url4_size,
			    (const char*)url6, url6_size);
			if (0 != error) {
				LOG_ERR(error, "upnp_ssdp_dev_if_add()");
				continue;
			}
		}
	}
	free(cfg_file_buf);
	upnp_ssdp_send_notify(g_data.upnp_ssdp);
    } /* Done with config. */

	if (0 == upnp_ssdp_root_dev_count(g_data.upnp_ssdp) ||
	    0 == upnp_ssdp_if_count(g_data.upnp_ssdp)) {
		LOG_INFO("no announce devices specified or no network ifaces, nothink to do...");
		goto exit_porgramm;
	}

	SigInstall();
	write_pid(g_data.cmd_line_data.pid_file_name); // Store pid to file

	set_user_and_group(g_data.cmd_line_data.pw_uid, g_data.cmd_line_data.pw_gid); // drop rights

	/* Receive and process packets. */
	thrp_threads_create(g_data.thrp, 1);
	thrp_thread_attach_first(g_data.thrp);
	thrp_shutdown_wait(g_data.thrp);

exit_porgramm:
	/* Deinitialization... */
	upnp_ssdp_destroy(g_data.upnp_ssdp);
	if (NULL != g_data.cmd_line_data.pid_file_name)
		unlink(g_data.cmd_line_data.pid_file_name); // Remove pid file

	thrp_destroy(g_data.thrp);
	LOG_INFO("exiting.");
	close(g_data.log_fd);

	return (0);
}

static void
SigInstall() {
	signal(SIGINT,	SigHandler);
	signal(SIGTERM,	SigHandler);
	signal(SIGKILL,	SigHandler);
	signal(SIGHUP,	SigHandler);
	signal(SIGUSR1,	SigHandler);
	signal(SIGUSR2,	SigHandler);
	signal(SIGPIPE,	SIG_IGN);
}

static void
SigHandler(int iSigNum) {

	switch (iSigNum) {
	case SIGINT:
	case SIGTERM:
	case SIGKILL:
		//exit(1);
		thrp_shutdown(g_data.thrp);
		break;
	case SIGHUP:
	case SIGUSR1:
	case SIGUSR2:
		break;
	}
}

